# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import json  # type: ignore

from google.auth.transport.requests import AuthorizedSession  # type: ignore
from google.auth import credentials as ga_credentials  # type: ignore
from google.api_core import exceptions as core_exceptions
from google.api_core import retry as retries
from google.api_core import rest_helpers
from google.api_core import rest_streaming
from google.api_core import gapic_v1
import cloudsdk.google.protobuf

from cloudsdk.google.protobuf import json_format
from google.api_core import operations_v1

from requests import __version__ as requests_version
import dataclasses
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union
import warnings


from googlecloudsdk.generated_clients.gapic_clients.aiplatform_v1beta1.types import semantic_governance_policy_service
from google.longrunning import operations_pb2  # type: ignore


from .rest_base import _BaseSemanticGovernancePolicyServiceRestTransport
from .base import DEFAULT_CLIENT_INFO as BASE_DEFAULT_CLIENT_INFO

try:
    OptionalRetry = Union[retries.Retry, gapic_v1.method._MethodDefault, None]
except AttributeError:  # pragma: NO COVER
    OptionalRetry = Union[retries.Retry, object, None]  # type: ignore

try:
    from google.api_core import client_logging  # type: ignore
    CLIENT_LOGGING_SUPPORTED = True  # pragma: NO COVER
except ImportError:  # pragma: NO COVER
    CLIENT_LOGGING_SUPPORTED = False

_LOGGER = logging.getLogger(__name__)

DEFAULT_CLIENT_INFO = gapic_v1.client_info.ClientInfo(
    gapic_version=BASE_DEFAULT_CLIENT_INFO.gapic_version,
    grpc_version=None,
    rest_version=f"requests@{requests_version}",
)

if hasattr(DEFAULT_CLIENT_INFO, "protobuf_runtime_version"):  # pragma: NO COVER
    DEFAULT_CLIENT_INFO.protobuf_runtime_version = cloudsdk.google.protobuf.__version__


class SemanticGovernancePolicyServiceRestInterceptor:
    """Interceptor for SemanticGovernancePolicyService.

    Interceptors are used to manipulate requests, request metadata, and responses
    in arbitrary ways.
    Example use cases include:
    * Logging
    * Verifying requests according to service or custom semantics
    * Stripping extraneous information from responses

    These use cases and more can be enabled by injecting an
    instance of a custom subclass when constructing the SemanticGovernancePolicyServiceRestTransport.

    .. code-block:: python
        class MyCustomSemanticGovernancePolicyServiceInterceptor(SemanticGovernancePolicyServiceRestInterceptor):
            def pre_create_semantic_governance_policy(self, request, metadata):
                logging.log(f"Received request: {request}")
                return request, metadata

            def post_create_semantic_governance_policy(self, response):
                logging.log(f"Received response: {response}")
                return response

            def pre_delete_semantic_governance_policy(self, request, metadata):
                logging.log(f"Received request: {request}")
                return request, metadata

            def post_delete_semantic_governance_policy(self, response):
                logging.log(f"Received response: {response}")
                return response

            def pre_get_semantic_governance_policy(self, request, metadata):
                logging.log(f"Received request: {request}")
                return request, metadata

            def post_get_semantic_governance_policy(self, response):
                logging.log(f"Received response: {response}")
                return response

            def pre_list_semantic_governance_policies(self, request, metadata):
                logging.log(f"Received request: {request}")
                return request, metadata

            def post_list_semantic_governance_policies(self, response):
                logging.log(f"Received response: {response}")
                return response

            def pre_update_semantic_governance_policy(self, request, metadata):
                logging.log(f"Received request: {request}")
                return request, metadata

            def post_update_semantic_governance_policy(self, response):
                logging.log(f"Received response: {response}")
                return response

        transport = SemanticGovernancePolicyServiceRestTransport(interceptor=MyCustomSemanticGovernancePolicyServiceInterceptor())
        client = SemanticGovernancePolicyServiceClient(transport=transport)


    """
    def pre_create_semantic_governance_policy(self, request: semantic_governance_policy_service.CreateSemanticGovernancePolicyRequest, metadata: Sequence[Tuple[str, Union[str, bytes]]]) -> Tuple[semantic_governance_policy_service.CreateSemanticGovernancePolicyRequest, Sequence[Tuple[str, Union[str, bytes]]]]:
        """Pre-rpc interceptor for create_semantic_governance_policy

        Override in a subclass to manipulate the request or metadata
        before they are sent to the SemanticGovernancePolicyService server.
        """
        return request, metadata

    def post_create_semantic_governance_policy(self, response: operations_pb2.Operation) -> operations_pb2.Operation:
        """Post-rpc interceptor for create_semantic_governance_policy

        DEPRECATED. Please use the `post_create_semantic_governance_policy_with_metadata`
        interceptor instead.

        Override in a subclass to read or manipulate the response
        after it is returned by the SemanticGovernancePolicyService server but before
        it is returned to user code. This `post_create_semantic_governance_policy` interceptor runs
        before the `post_create_semantic_governance_policy_with_metadata` interceptor.
        """
        return response

    def post_create_semantic_governance_policy_with_metadata(self, response: operations_pb2.Operation, metadata: Sequence[Tuple[str, Union[str, bytes]]]) -> Tuple[operations_pb2.Operation, Sequence[Tuple[str, Union[str, bytes]]]]:
        """Post-rpc interceptor for create_semantic_governance_policy

        Override in a subclass to read or manipulate the response or metadata after it
        is returned by the SemanticGovernancePolicyService server but before it is returned to user code.

        We recommend only using this `post_create_semantic_governance_policy_with_metadata`
        interceptor in new development instead of the `post_create_semantic_governance_policy` interceptor.
        When both interceptors are used, this `post_create_semantic_governance_policy_with_metadata` interceptor runs after the
        `post_create_semantic_governance_policy` interceptor. The (possibly modified) response returned by
        `post_create_semantic_governance_policy` will be passed to
        `post_create_semantic_governance_policy_with_metadata`.
        """
        return response, metadata

    def pre_delete_semantic_governance_policy(self, request: semantic_governance_policy_service.DeleteSemanticGovernancePolicyRequest, metadata: Sequence[Tuple[str, Union[str, bytes]]]) -> Tuple[semantic_governance_policy_service.DeleteSemanticGovernancePolicyRequest, Sequence[Tuple[str, Union[str, bytes]]]]:
        """Pre-rpc interceptor for delete_semantic_governance_policy

        Override in a subclass to manipulate the request or metadata
        before they are sent to the SemanticGovernancePolicyService server.
        """
        return request, metadata

    def post_delete_semantic_governance_policy(self, response: operations_pb2.Operation) -> operations_pb2.Operation:
        """Post-rpc interceptor for delete_semantic_governance_policy

        DEPRECATED. Please use the `post_delete_semantic_governance_policy_with_metadata`
        interceptor instead.

        Override in a subclass to read or manipulate the response
        after it is returned by the SemanticGovernancePolicyService server but before
        it is returned to user code. This `post_delete_semantic_governance_policy` interceptor runs
        before the `post_delete_semantic_governance_policy_with_metadata` interceptor.
        """
        return response

    def post_delete_semantic_governance_policy_with_metadata(self, response: operations_pb2.Operation, metadata: Sequence[Tuple[str, Union[str, bytes]]]) -> Tuple[operations_pb2.Operation, Sequence[Tuple[str, Union[str, bytes]]]]:
        """Post-rpc interceptor for delete_semantic_governance_policy

        Override in a subclass to read or manipulate the response or metadata after it
        is returned by the SemanticGovernancePolicyService server but before it is returned to user code.

        We recommend only using this `post_delete_semantic_governance_policy_with_metadata`
        interceptor in new development instead of the `post_delete_semantic_governance_policy` interceptor.
        When both interceptors are used, this `post_delete_semantic_governance_policy_with_metadata` interceptor runs after the
        `post_delete_semantic_governance_policy` interceptor. The (possibly modified) response returned by
        `post_delete_semantic_governance_policy` will be passed to
        `post_delete_semantic_governance_policy_with_metadata`.
        """
        return response, metadata

    def pre_get_semantic_governance_policy(self, request: semantic_governance_policy_service.GetSemanticGovernancePolicyRequest, metadata: Sequence[Tuple[str, Union[str, bytes]]]) -> Tuple[semantic_governance_policy_service.GetSemanticGovernancePolicyRequest, Sequence[Tuple[str, Union[str, bytes]]]]:
        """Pre-rpc interceptor for get_semantic_governance_policy

        Override in a subclass to manipulate the request or metadata
        before they are sent to the SemanticGovernancePolicyService server.
        """
        return request, metadata

    def post_get_semantic_governance_policy(self, response: semantic_governance_policy_service.SemanticGovernancePolicy) -> semantic_governance_policy_service.SemanticGovernancePolicy:
        """Post-rpc interceptor for get_semantic_governance_policy

        DEPRECATED. Please use the `post_get_semantic_governance_policy_with_metadata`
        interceptor instead.

        Override in a subclass to read or manipulate the response
        after it is returned by the SemanticGovernancePolicyService server but before
        it is returned to user code. This `post_get_semantic_governance_policy` interceptor runs
        before the `post_get_semantic_governance_policy_with_metadata` interceptor.
        """
        return response

    def post_get_semantic_governance_policy_with_metadata(self, response: semantic_governance_policy_service.SemanticGovernancePolicy, metadata: Sequence[Tuple[str, Union[str, bytes]]]) -> Tuple[semantic_governance_policy_service.SemanticGovernancePolicy, Sequence[Tuple[str, Union[str, bytes]]]]:
        """Post-rpc interceptor for get_semantic_governance_policy

        Override in a subclass to read or manipulate the response or metadata after it
        is returned by the SemanticGovernancePolicyService server but before it is returned to user code.

        We recommend only using this `post_get_semantic_governance_policy_with_metadata`
        interceptor in new development instead of the `post_get_semantic_governance_policy` interceptor.
        When both interceptors are used, this `post_get_semantic_governance_policy_with_metadata` interceptor runs after the
        `post_get_semantic_governance_policy` interceptor. The (possibly modified) response returned by
        `post_get_semantic_governance_policy` will be passed to
        `post_get_semantic_governance_policy_with_metadata`.
        """
        return response, metadata

    def pre_list_semantic_governance_policies(self, request: semantic_governance_policy_service.ListSemanticGovernancePoliciesRequest, metadata: Sequence[Tuple[str, Union[str, bytes]]]) -> Tuple[semantic_governance_policy_service.ListSemanticGovernancePoliciesRequest, Sequence[Tuple[str, Union[str, bytes]]]]:
        """Pre-rpc interceptor for list_semantic_governance_policies

        Override in a subclass to manipulate the request or metadata
        before they are sent to the SemanticGovernancePolicyService server.
        """
        return request, metadata

    def post_list_semantic_governance_policies(self, response: semantic_governance_policy_service.ListSemanticGovernancePoliciesResponse) -> semantic_governance_policy_service.ListSemanticGovernancePoliciesResponse:
        """Post-rpc interceptor for list_semantic_governance_policies

        DEPRECATED. Please use the `post_list_semantic_governance_policies_with_metadata`
        interceptor instead.

        Override in a subclass to read or manipulate the response
        after it is returned by the SemanticGovernancePolicyService server but before
        it is returned to user code. This `post_list_semantic_governance_policies` interceptor runs
        before the `post_list_semantic_governance_policies_with_metadata` interceptor.
        """
        return response

    def post_list_semantic_governance_policies_with_metadata(self, response: semantic_governance_policy_service.ListSemanticGovernancePoliciesResponse, metadata: Sequence[Tuple[str, Union[str, bytes]]]) -> Tuple[semantic_governance_policy_service.ListSemanticGovernancePoliciesResponse, Sequence[Tuple[str, Union[str, bytes]]]]:
        """Post-rpc interceptor for list_semantic_governance_policies

        Override in a subclass to read or manipulate the response or metadata after it
        is returned by the SemanticGovernancePolicyService server but before it is returned to user code.

        We recommend only using this `post_list_semantic_governance_policies_with_metadata`
        interceptor in new development instead of the `post_list_semantic_governance_policies` interceptor.
        When both interceptors are used, this `post_list_semantic_governance_policies_with_metadata` interceptor runs after the
        `post_list_semantic_governance_policies` interceptor. The (possibly modified) response returned by
        `post_list_semantic_governance_policies` will be passed to
        `post_list_semantic_governance_policies_with_metadata`.
        """
        return response, metadata

    def pre_update_semantic_governance_policy(self, request: semantic_governance_policy_service.UpdateSemanticGovernancePolicyRequest, metadata: Sequence[Tuple[str, Union[str, bytes]]]) -> Tuple[semantic_governance_policy_service.UpdateSemanticGovernancePolicyRequest, Sequence[Tuple[str, Union[str, bytes]]]]:
        """Pre-rpc interceptor for update_semantic_governance_policy

        Override in a subclass to manipulate the request or metadata
        before they are sent to the SemanticGovernancePolicyService server.
        """
        return request, metadata

    def post_update_semantic_governance_policy(self, response: operations_pb2.Operation) -> operations_pb2.Operation:
        """Post-rpc interceptor for update_semantic_governance_policy

        DEPRECATED. Please use the `post_update_semantic_governance_policy_with_metadata`
        interceptor instead.

        Override in a subclass to read or manipulate the response
        after it is returned by the SemanticGovernancePolicyService server but before
        it is returned to user code. This `post_update_semantic_governance_policy` interceptor runs
        before the `post_update_semantic_governance_policy_with_metadata` interceptor.
        """
        return response

    def post_update_semantic_governance_policy_with_metadata(self, response: operations_pb2.Operation, metadata: Sequence[Tuple[str, Union[str, bytes]]]) -> Tuple[operations_pb2.Operation, Sequence[Tuple[str, Union[str, bytes]]]]:
        """Post-rpc interceptor for update_semantic_governance_policy

        Override in a subclass to read or manipulate the response or metadata after it
        is returned by the SemanticGovernancePolicyService server but before it is returned to user code.

        We recommend only using this `post_update_semantic_governance_policy_with_metadata`
        interceptor in new development instead of the `post_update_semantic_governance_policy` interceptor.
        When both interceptors are used, this `post_update_semantic_governance_policy_with_metadata` interceptor runs after the
        `post_update_semantic_governance_policy` interceptor. The (possibly modified) response returned by
        `post_update_semantic_governance_policy` will be passed to
        `post_update_semantic_governance_policy_with_metadata`.
        """
        return response, metadata


@dataclasses.dataclass
class SemanticGovernancePolicyServiceRestStub:
    _session: AuthorizedSession
    _host: str
    _interceptor: SemanticGovernancePolicyServiceRestInterceptor


class SemanticGovernancePolicyServiceRestTransport(_BaseSemanticGovernancePolicyServiceRestTransport):
    """REST backend synchronous transport for SemanticGovernancePolicyService.

    Manages SemanticGovernancePolicies.
    A SemanticGovernancePolicy is a resource that represents a
    user-specified policy applied to an Agent and optionally a
    specific Tool within that Agent.

    This class defines the same methods as the primary client, so the
    primary client can load the underlying transport implementation
    and call it.

    It sends JSON representations of protocol buffers over HTTP/1.1
    """

    def __init__(self, *,
            host: str = 'aiplatform.googleapis.com',
            credentials: Optional[ga_credentials.Credentials] = None,
            credentials_file: Optional[str] = None,
            scopes: Optional[Sequence[str]] = None,
            client_cert_source_for_mtls: Optional[Callable[[
                ], Tuple[bytes, bytes]]] = None,
            quota_project_id: Optional[str] = None,
            client_info: gapic_v1.client_info.ClientInfo = DEFAULT_CLIENT_INFO,
            always_use_jwt_access: Optional[bool] = False,
            url_scheme: str = 'https',
            interceptor: Optional[SemanticGovernancePolicyServiceRestInterceptor] = None,
            api_audience: Optional[str] = None,
            ) -> None:
        """Instantiate the transport.

       NOTE: This REST transport functionality is currently in a beta
       state (preview). We welcome your feedback via a GitHub issue in
       this library's repository. Thank you!

        Args:
            host (Optional[str]):
                 The hostname to connect to (default: 'aiplatform.googleapis.com').
            credentials (Optional[google.auth.credentials.Credentials]): The
                authorization credentials to attach to requests. These
                credentials identify the application to the service; if none
                are specified, the client will attempt to ascertain the
                credentials from the environment.

            credentials_file (Optional[str]): Deprecated. A file with credentials that can
                be loaded with :func:`google.auth.load_credentials_from_file`.
                This argument is ignored if ``channel`` is provided. This argument will be
                removed in the next major version of this library.
            scopes (Optional(Sequence[str])): A list of scopes. This argument is
                ignored if ``channel`` is provided.
            client_cert_source_for_mtls (Callable[[], Tuple[bytes, bytes]]): Client
                certificate to configure mutual TLS HTTP channel. It is ignored
                if ``channel`` is provided.
            quota_project_id (Optional[str]): An optional project to use for billing
                and quota.
            client_info (google.api_core.gapic_v1.client_info.ClientInfo):
                The client info used to send a user-agent string along with
                API requests. If ``None``, then default info will be used.
                Generally, you only need to set this if you are developing
                your own client library.
            always_use_jwt_access (Optional[bool]): Whether self signed JWT should
                be used for service account credentials.
            url_scheme: the protocol scheme for the API endpoint.  Normally
                "https", but for testing or local servers,
                "http" can be specified.
        """
        # Run the base constructor
        # TODO(yon-mg): resolve other ctor params i.e. scopes, quota, etc.
        # TODO: When custom host (api_endpoint) is set, `scopes` must *also* be set on the
        # credentials object
        super().__init__(
            host=host,
            credentials=credentials,
            client_info=client_info,
            always_use_jwt_access=always_use_jwt_access,
            url_scheme=url_scheme,
            api_audience=api_audience
        )
        self._session = AuthorizedSession(
            self._credentials, default_host=self.DEFAULT_HOST)
        self._operations_client: Optional[operations_v1.AbstractOperationsClient] = None
        if client_cert_source_for_mtls:
            self._session.configure_mtls_channel(client_cert_source_for_mtls)
        self._interceptor = interceptor or SemanticGovernancePolicyServiceRestInterceptor()
        self._prep_wrapped_messages(client_info)

    @property
    def operations_client(self) -> operations_v1.AbstractOperationsClient:
        """Create the client designed to process long-running operations.

        This property caches on the instance; repeated calls return the same
        client.
        """
        # Only create a new client if we do not already have one.
        if self._operations_client is None:
            http_options: Dict[str, List[Dict[str, str]]] = {
            }

            rest_transport = operations_v1.OperationsRestTransport(
                    host=self._host,
                    # use the credentials which are saved
                    credentials=self._credentials,
                    scopes=self._scopes,
                    http_options=http_options,
                    path_prefix="v1beta1")

            self._operations_client = operations_v1.AbstractOperationsClient(transport=rest_transport)

        # Return the client from cache.
        return self._operations_client

    class _CreateSemanticGovernancePolicy(_BaseSemanticGovernancePolicyServiceRestTransport._BaseCreateSemanticGovernancePolicy, SemanticGovernancePolicyServiceRestStub):
        def __hash__(self):
            return hash("SemanticGovernancePolicyServiceRestTransport.CreateSemanticGovernancePolicy")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None):

            uri = transcoded_request['uri']
            method = transcoded_request['method']
            headers = dict(metadata)
            headers['Content-Type'] = 'application/json'
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
                data=body,
                )
            return response

        def __call__(self,
                request: semantic_governance_policy_service.CreateSemanticGovernancePolicyRequest, *,
                retry: OptionalRetry=gapic_v1.method.DEFAULT,
                timeout: Optional[float]=None,
                metadata: Sequence[Tuple[str, Union[str, bytes]]]=(),
                ) -> operations_pb2.Operation:
            r"""Call the create semantic
        governance policy method over HTTP.

            Args:
                request (~.semantic_governance_policy_service.CreateSemanticGovernancePolicyRequest):
                    The request object. Request message for
                SemanticGovernancePolicyService.CreateSemanticGovernancePolicy.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, Union[str, bytes]]]): Key/value pairs which should be
                    sent along with the request as metadata. Normally, each value must be of type `str`,
                    but for metadata keys ending with the suffix `-bin`, the corresponding values must
                    be of type `bytes`.

            Returns:
                ~.operations_pb2.Operation:
                    This resource represents a
                long-running operation that is the
                result of a network API call.

            """

            http_options = _BaseSemanticGovernancePolicyServiceRestTransport._BaseCreateSemanticGovernancePolicy._get_http_options()

            request, metadata = self._interceptor.pre_create_semantic_governance_policy(request, metadata)
            transcoded_request = _BaseSemanticGovernancePolicyServiceRestTransport._BaseCreateSemanticGovernancePolicy._get_transcoded_request(http_options, request)

            body = _BaseSemanticGovernancePolicyServiceRestTransport._BaseCreateSemanticGovernancePolicy._get_request_body_json(transcoded_request)

            # Jsonify the query params
            query_params = _BaseSemanticGovernancePolicyServiceRestTransport._BaseCreateSemanticGovernancePolicy._get_query_params_json(transcoded_request)

            if CLIENT_LOGGING_SUPPORTED and _LOGGER.isEnabledFor(logging.DEBUG):  # pragma: NO COVER
                request_url = "{host}{uri}".format(host=self._host, uri=transcoded_request['uri'])
                method = transcoded_request['method']
                try:
                    request_payload = json_format.MessageToJson(request)
                except:
                    request_payload = None
                http_request = {
                  "payload": request_payload,
                  "requestMethod": method,
                  "requestUrl": request_url,
                  "headers": dict(metadata),
                }
                _LOGGER.debug(
                    f"Sending request for google.cloud.aiplatform_v1beta1.SemanticGovernancePolicyServiceClient.CreateSemanticGovernancePolicy",
                    extra = {
                        "serviceName": "google.cloud.aiplatform.v1beta1.SemanticGovernancePolicyService",
                        "rpcName": "CreateSemanticGovernancePolicy",
                        "httpRequest": http_request,
                        "metadata": http_request["headers"],
                    },
                )

            # Send the request
            response = SemanticGovernancePolicyServiceRestTransport._CreateSemanticGovernancePolicy._get_response(self._host, metadata, query_params, self._session, timeout, transcoded_request, body)

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            # Return the response
            resp = operations_pb2.Operation()
            json_format.Parse(response.content, resp, ignore_unknown_fields=True)

            resp = self._interceptor.post_create_semantic_governance_policy(resp)
            response_metadata = [(k, str(v)) for k, v in response.headers.items()]
            resp, _ = self._interceptor.post_create_semantic_governance_policy_with_metadata(resp, response_metadata)
            if CLIENT_LOGGING_SUPPORTED and _LOGGER.isEnabledFor(logging.DEBUG):  # pragma: NO COVER
                try:
                    response_payload = json_format.MessageToJson(resp)
                except:
                    response_payload = None
                http_response = {
                "payload": response_payload,
                "headers":  dict(response.headers),
                "status": response.status_code,
                }
                _LOGGER.debug(
                    "Received response for google.cloud.aiplatform_v1beta1.SemanticGovernancePolicyServiceClient.create_semantic_governance_policy",
                    extra = {
                        "serviceName": "google.cloud.aiplatform.v1beta1.SemanticGovernancePolicyService",
                        "rpcName": "CreateSemanticGovernancePolicy",
                        "metadata": http_response["headers"],
                        "httpResponse": http_response,
                    },
                )
            return resp

    class _DeleteSemanticGovernancePolicy(_BaseSemanticGovernancePolicyServiceRestTransport._BaseDeleteSemanticGovernancePolicy, SemanticGovernancePolicyServiceRestStub):
        def __hash__(self):
            return hash("SemanticGovernancePolicyServiceRestTransport.DeleteSemanticGovernancePolicy")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None):

            uri = transcoded_request['uri']
            method = transcoded_request['method']
            headers = dict(metadata)
            headers['Content-Type'] = 'application/json'
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
                )
            return response

        def __call__(self,
                request: semantic_governance_policy_service.DeleteSemanticGovernancePolicyRequest, *,
                retry: OptionalRetry=gapic_v1.method.DEFAULT,
                timeout: Optional[float]=None,
                metadata: Sequence[Tuple[str, Union[str, bytes]]]=(),
                ) -> operations_pb2.Operation:
            r"""Call the delete semantic
        governance policy method over HTTP.

            Args:
                request (~.semantic_governance_policy_service.DeleteSemanticGovernancePolicyRequest):
                    The request object. Request message for
                SemanticGovernancePolicyService.DeleteSemanticGovernancePolicy.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, Union[str, bytes]]]): Key/value pairs which should be
                    sent along with the request as metadata. Normally, each value must be of type `str`,
                    but for metadata keys ending with the suffix `-bin`, the corresponding values must
                    be of type `bytes`.

            Returns:
                ~.operations_pb2.Operation:
                    This resource represents a
                long-running operation that is the
                result of a network API call.

            """

            http_options = _BaseSemanticGovernancePolicyServiceRestTransport._BaseDeleteSemanticGovernancePolicy._get_http_options()

            request, metadata = self._interceptor.pre_delete_semantic_governance_policy(request, metadata)
            transcoded_request = _BaseSemanticGovernancePolicyServiceRestTransport._BaseDeleteSemanticGovernancePolicy._get_transcoded_request(http_options, request)

            # Jsonify the query params
            query_params = _BaseSemanticGovernancePolicyServiceRestTransport._BaseDeleteSemanticGovernancePolicy._get_query_params_json(transcoded_request)

            if CLIENT_LOGGING_SUPPORTED and _LOGGER.isEnabledFor(logging.DEBUG):  # pragma: NO COVER
                request_url = "{host}{uri}".format(host=self._host, uri=transcoded_request['uri'])
                method = transcoded_request['method']
                try:
                    request_payload = json_format.MessageToJson(request)
                except:
                    request_payload = None
                http_request = {
                  "payload": request_payload,
                  "requestMethod": method,
                  "requestUrl": request_url,
                  "headers": dict(metadata),
                }
                _LOGGER.debug(
                    f"Sending request for google.cloud.aiplatform_v1beta1.SemanticGovernancePolicyServiceClient.DeleteSemanticGovernancePolicy",
                    extra = {
                        "serviceName": "google.cloud.aiplatform.v1beta1.SemanticGovernancePolicyService",
                        "rpcName": "DeleteSemanticGovernancePolicy",
                        "httpRequest": http_request,
                        "metadata": http_request["headers"],
                    },
                )

            # Send the request
            response = SemanticGovernancePolicyServiceRestTransport._DeleteSemanticGovernancePolicy._get_response(self._host, metadata, query_params, self._session, timeout, transcoded_request)

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            # Return the response
            resp = operations_pb2.Operation()
            json_format.Parse(response.content, resp, ignore_unknown_fields=True)

            resp = self._interceptor.post_delete_semantic_governance_policy(resp)
            response_metadata = [(k, str(v)) for k, v in response.headers.items()]
            resp, _ = self._interceptor.post_delete_semantic_governance_policy_with_metadata(resp, response_metadata)
            if CLIENT_LOGGING_SUPPORTED and _LOGGER.isEnabledFor(logging.DEBUG):  # pragma: NO COVER
                try:
                    response_payload = json_format.MessageToJson(resp)
                except:
                    response_payload = None
                http_response = {
                "payload": response_payload,
                "headers":  dict(response.headers),
                "status": response.status_code,
                }
                _LOGGER.debug(
                    "Received response for google.cloud.aiplatform_v1beta1.SemanticGovernancePolicyServiceClient.delete_semantic_governance_policy",
                    extra = {
                        "serviceName": "google.cloud.aiplatform.v1beta1.SemanticGovernancePolicyService",
                        "rpcName": "DeleteSemanticGovernancePolicy",
                        "metadata": http_response["headers"],
                        "httpResponse": http_response,
                    },
                )
            return resp

    class _GetSemanticGovernancePolicy(_BaseSemanticGovernancePolicyServiceRestTransport._BaseGetSemanticGovernancePolicy, SemanticGovernancePolicyServiceRestStub):
        def __hash__(self):
            return hash("SemanticGovernancePolicyServiceRestTransport.GetSemanticGovernancePolicy")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None):

            uri = transcoded_request['uri']
            method = transcoded_request['method']
            headers = dict(metadata)
            headers['Content-Type'] = 'application/json'
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
                )
            return response

        def __call__(self,
                request: semantic_governance_policy_service.GetSemanticGovernancePolicyRequest, *,
                retry: OptionalRetry=gapic_v1.method.DEFAULT,
                timeout: Optional[float]=None,
                metadata: Sequence[Tuple[str, Union[str, bytes]]]=(),
                ) -> semantic_governance_policy_service.SemanticGovernancePolicy:
            r"""Call the get semantic governance
        policy method over HTTP.

            Args:
                request (~.semantic_governance_policy_service.GetSemanticGovernancePolicyRequest):
                    The request object. Request message for
                SemanticGovernancePolicyService.GetSemanticGovernancePolicy.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, Union[str, bytes]]]): Key/value pairs which should be
                    sent along with the request as metadata. Normally, each value must be of type `str`,
                    but for metadata keys ending with the suffix `-bin`, the corresponding values must
                    be of type `bytes`.

            Returns:
                ~.semantic_governance_policy_service.SemanticGovernancePolicy:
                    Represents a governance policy
                applied to a specific Agent and
                optionally a specific Tool within that
                Agent.

            """

            http_options = _BaseSemanticGovernancePolicyServiceRestTransport._BaseGetSemanticGovernancePolicy._get_http_options()

            request, metadata = self._interceptor.pre_get_semantic_governance_policy(request, metadata)
            transcoded_request = _BaseSemanticGovernancePolicyServiceRestTransport._BaseGetSemanticGovernancePolicy._get_transcoded_request(http_options, request)

            # Jsonify the query params
            query_params = _BaseSemanticGovernancePolicyServiceRestTransport._BaseGetSemanticGovernancePolicy._get_query_params_json(transcoded_request)

            if CLIENT_LOGGING_SUPPORTED and _LOGGER.isEnabledFor(logging.DEBUG):  # pragma: NO COVER
                request_url = "{host}{uri}".format(host=self._host, uri=transcoded_request['uri'])
                method = transcoded_request['method']
                try:
                    request_payload = type(request).to_json(request)
                except:
                    request_payload = None
                http_request = {
                  "payload": request_payload,
                  "requestMethod": method,
                  "requestUrl": request_url,
                  "headers": dict(metadata),
                }
                _LOGGER.debug(
                    f"Sending request for google.cloud.aiplatform_v1beta1.SemanticGovernancePolicyServiceClient.GetSemanticGovernancePolicy",
                    extra = {
                        "serviceName": "google.cloud.aiplatform.v1beta1.SemanticGovernancePolicyService",
                        "rpcName": "GetSemanticGovernancePolicy",
                        "httpRequest": http_request,
                        "metadata": http_request["headers"],
                    },
                )

            # Send the request
            response = SemanticGovernancePolicyServiceRestTransport._GetSemanticGovernancePolicy._get_response(self._host, metadata, query_params, self._session, timeout, transcoded_request)

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            # Return the response
            resp = semantic_governance_policy_service.SemanticGovernancePolicy()
            pb_resp = semantic_governance_policy_service.SemanticGovernancePolicy.pb(resp)

            json_format.Parse(response.content, pb_resp, ignore_unknown_fields=True)

            resp = self._interceptor.post_get_semantic_governance_policy(resp)
            response_metadata = [(k, str(v)) for k, v in response.headers.items()]
            resp, _ = self._interceptor.post_get_semantic_governance_policy_with_metadata(resp, response_metadata)
            if CLIENT_LOGGING_SUPPORTED and _LOGGER.isEnabledFor(logging.DEBUG):  # pragma: NO COVER
                try:
                    response_payload = semantic_governance_policy_service.SemanticGovernancePolicy.to_json(response)
                except:
                    response_payload = None
                http_response = {
                "payload": response_payload,
                "headers":  dict(response.headers),
                "status": response.status_code,
                }
                _LOGGER.debug(
                    "Received response for google.cloud.aiplatform_v1beta1.SemanticGovernancePolicyServiceClient.get_semantic_governance_policy",
                    extra = {
                        "serviceName": "google.cloud.aiplatform.v1beta1.SemanticGovernancePolicyService",
                        "rpcName": "GetSemanticGovernancePolicy",
                        "metadata": http_response["headers"],
                        "httpResponse": http_response,
                    },
                )
            return resp

    class _ListSemanticGovernancePolicies(_BaseSemanticGovernancePolicyServiceRestTransport._BaseListSemanticGovernancePolicies, SemanticGovernancePolicyServiceRestStub):
        def __hash__(self):
            return hash("SemanticGovernancePolicyServiceRestTransport.ListSemanticGovernancePolicies")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None):

            uri = transcoded_request['uri']
            method = transcoded_request['method']
            headers = dict(metadata)
            headers['Content-Type'] = 'application/json'
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
                )
            return response

        def __call__(self,
                request: semantic_governance_policy_service.ListSemanticGovernancePoliciesRequest, *,
                retry: OptionalRetry=gapic_v1.method.DEFAULT,
                timeout: Optional[float]=None,
                metadata: Sequence[Tuple[str, Union[str, bytes]]]=(),
                ) -> semantic_governance_policy_service.ListSemanticGovernancePoliciesResponse:
            r"""Call the list semantic governance
        policies method over HTTP.

            Args:
                request (~.semantic_governance_policy_service.ListSemanticGovernancePoliciesRequest):
                    The request object. Request message for
                SemanticGovernancePolicyService.ListSemanticGovernancePolicies.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, Union[str, bytes]]]): Key/value pairs which should be
                    sent along with the request as metadata. Normally, each value must be of type `str`,
                    but for metadata keys ending with the suffix `-bin`, the corresponding values must
                    be of type `bytes`.

            Returns:
                ~.semantic_governance_policy_service.ListSemanticGovernancePoliciesResponse:
                    Response message for
                SemanticGovernancePolicyService.ListSemanticGovernancePolicies.

            """

            http_options = _BaseSemanticGovernancePolicyServiceRestTransport._BaseListSemanticGovernancePolicies._get_http_options()

            request, metadata = self._interceptor.pre_list_semantic_governance_policies(request, metadata)
            transcoded_request = _BaseSemanticGovernancePolicyServiceRestTransport._BaseListSemanticGovernancePolicies._get_transcoded_request(http_options, request)

            # Jsonify the query params
            query_params = _BaseSemanticGovernancePolicyServiceRestTransport._BaseListSemanticGovernancePolicies._get_query_params_json(transcoded_request)

            if CLIENT_LOGGING_SUPPORTED and _LOGGER.isEnabledFor(logging.DEBUG):  # pragma: NO COVER
                request_url = "{host}{uri}".format(host=self._host, uri=transcoded_request['uri'])
                method = transcoded_request['method']
                try:
                    request_payload = type(request).to_json(request)
                except:
                    request_payload = None
                http_request = {
                  "payload": request_payload,
                  "requestMethod": method,
                  "requestUrl": request_url,
                  "headers": dict(metadata),
                }
                _LOGGER.debug(
                    f"Sending request for google.cloud.aiplatform_v1beta1.SemanticGovernancePolicyServiceClient.ListSemanticGovernancePolicies",
                    extra = {
                        "serviceName": "google.cloud.aiplatform.v1beta1.SemanticGovernancePolicyService",
                        "rpcName": "ListSemanticGovernancePolicies",
                        "httpRequest": http_request,
                        "metadata": http_request["headers"],
                    },
                )

            # Send the request
            response = SemanticGovernancePolicyServiceRestTransport._ListSemanticGovernancePolicies._get_response(self._host, metadata, query_params, self._session, timeout, transcoded_request)

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            # Return the response
            resp = semantic_governance_policy_service.ListSemanticGovernancePoliciesResponse()
            pb_resp = semantic_governance_policy_service.ListSemanticGovernancePoliciesResponse.pb(resp)

            json_format.Parse(response.content, pb_resp, ignore_unknown_fields=True)

            resp = self._interceptor.post_list_semantic_governance_policies(resp)
            response_metadata = [(k, str(v)) for k, v in response.headers.items()]
            resp, _ = self._interceptor.post_list_semantic_governance_policies_with_metadata(resp, response_metadata)
            if CLIENT_LOGGING_SUPPORTED and _LOGGER.isEnabledFor(logging.DEBUG):  # pragma: NO COVER
                try:
                    response_payload = semantic_governance_policy_service.ListSemanticGovernancePoliciesResponse.to_json(response)
                except:
                    response_payload = None
                http_response = {
                "payload": response_payload,
                "headers":  dict(response.headers),
                "status": response.status_code,
                }
                _LOGGER.debug(
                    "Received response for google.cloud.aiplatform_v1beta1.SemanticGovernancePolicyServiceClient.list_semantic_governance_policies",
                    extra = {
                        "serviceName": "google.cloud.aiplatform.v1beta1.SemanticGovernancePolicyService",
                        "rpcName": "ListSemanticGovernancePolicies",
                        "metadata": http_response["headers"],
                        "httpResponse": http_response,
                    },
                )
            return resp

    class _UpdateSemanticGovernancePolicy(_BaseSemanticGovernancePolicyServiceRestTransport._BaseUpdateSemanticGovernancePolicy, SemanticGovernancePolicyServiceRestStub):
        def __hash__(self):
            return hash("SemanticGovernancePolicyServiceRestTransport.UpdateSemanticGovernancePolicy")

        @staticmethod
        def _get_response(
            host,
            metadata,
            query_params,
            session,
            timeout,
            transcoded_request,
            body=None):

            uri = transcoded_request['uri']
            method = transcoded_request['method']
            headers = dict(metadata)
            headers['Content-Type'] = 'application/json'
            response = getattr(session, method)(
                "{host}{uri}".format(host=host, uri=uri),
                timeout=timeout,
                headers=headers,
                params=rest_helpers.flatten_query_params(query_params, strict=True),
                data=body,
                )
            return response

        def __call__(self,
                request: semantic_governance_policy_service.UpdateSemanticGovernancePolicyRequest, *,
                retry: OptionalRetry=gapic_v1.method.DEFAULT,
                timeout: Optional[float]=None,
                metadata: Sequence[Tuple[str, Union[str, bytes]]]=(),
                ) -> operations_pb2.Operation:
            r"""Call the update semantic
        governance policy method over HTTP.

            Args:
                request (~.semantic_governance_policy_service.UpdateSemanticGovernancePolicyRequest):
                    The request object. Request message for
                SemanticGovernancePolicyService.UpdateSemanticGovernancePolicy.
                retry (google.api_core.retry.Retry): Designation of what errors, if any,
                    should be retried.
                timeout (float): The timeout for this request.
                metadata (Sequence[Tuple[str, Union[str, bytes]]]): Key/value pairs which should be
                    sent along with the request as metadata. Normally, each value must be of type `str`,
                    but for metadata keys ending with the suffix `-bin`, the corresponding values must
                    be of type `bytes`.

            Returns:
                ~.operations_pb2.Operation:
                    This resource represents a
                long-running operation that is the
                result of a network API call.

            """

            http_options = _BaseSemanticGovernancePolicyServiceRestTransport._BaseUpdateSemanticGovernancePolicy._get_http_options()

            request, metadata = self._interceptor.pre_update_semantic_governance_policy(request, metadata)
            transcoded_request = _BaseSemanticGovernancePolicyServiceRestTransport._BaseUpdateSemanticGovernancePolicy._get_transcoded_request(http_options, request)

            body = _BaseSemanticGovernancePolicyServiceRestTransport._BaseUpdateSemanticGovernancePolicy._get_request_body_json(transcoded_request)

            # Jsonify the query params
            query_params = _BaseSemanticGovernancePolicyServiceRestTransport._BaseUpdateSemanticGovernancePolicy._get_query_params_json(transcoded_request)

            if CLIENT_LOGGING_SUPPORTED and _LOGGER.isEnabledFor(logging.DEBUG):  # pragma: NO COVER
                request_url = "{host}{uri}".format(host=self._host, uri=transcoded_request['uri'])
                method = transcoded_request['method']
                try:
                    request_payload = json_format.MessageToJson(request)
                except:
                    request_payload = None
                http_request = {
                  "payload": request_payload,
                  "requestMethod": method,
                  "requestUrl": request_url,
                  "headers": dict(metadata),
                }
                _LOGGER.debug(
                    f"Sending request for google.cloud.aiplatform_v1beta1.SemanticGovernancePolicyServiceClient.UpdateSemanticGovernancePolicy",
                    extra = {
                        "serviceName": "google.cloud.aiplatform.v1beta1.SemanticGovernancePolicyService",
                        "rpcName": "UpdateSemanticGovernancePolicy",
                        "httpRequest": http_request,
                        "metadata": http_request["headers"],
                    },
                )

            # Send the request
            response = SemanticGovernancePolicyServiceRestTransport._UpdateSemanticGovernancePolicy._get_response(self._host, metadata, query_params, self._session, timeout, transcoded_request, body)

            # In case of error, raise the appropriate core_exceptions.GoogleAPICallError exception
            # subclass.
            if response.status_code >= 400:
                raise core_exceptions.from_http_response(response)

            # Return the response
            resp = operations_pb2.Operation()
            json_format.Parse(response.content, resp, ignore_unknown_fields=True)

            resp = self._interceptor.post_update_semantic_governance_policy(resp)
            response_metadata = [(k, str(v)) for k, v in response.headers.items()]
            resp, _ = self._interceptor.post_update_semantic_governance_policy_with_metadata(resp, response_metadata)
            if CLIENT_LOGGING_SUPPORTED and _LOGGER.isEnabledFor(logging.DEBUG):  # pragma: NO COVER
                try:
                    response_payload = json_format.MessageToJson(resp)
                except:
                    response_payload = None
                http_response = {
                "payload": response_payload,
                "headers":  dict(response.headers),
                "status": response.status_code,
                }
                _LOGGER.debug(
                    "Received response for google.cloud.aiplatform_v1beta1.SemanticGovernancePolicyServiceClient.update_semantic_governance_policy",
                    extra = {
                        "serviceName": "google.cloud.aiplatform.v1beta1.SemanticGovernancePolicyService",
                        "rpcName": "UpdateSemanticGovernancePolicy",
                        "metadata": http_response["headers"],
                        "httpResponse": http_response,
                    },
                )
            return resp

    @property
    def create_semantic_governance_policy(self) -> Callable[
            [semantic_governance_policy_service.CreateSemanticGovernancePolicyRequest],
            operations_pb2.Operation]:
        # The return type is fine, but mypy isn't sophisticated enough to determine what's going on here.
        # In C++ this would require a dynamic_cast
        return self._CreateSemanticGovernancePolicy(self._session, self._host, self._interceptor) # type: ignore

    @property
    def delete_semantic_governance_policy(self) -> Callable[
            [semantic_governance_policy_service.DeleteSemanticGovernancePolicyRequest],
            operations_pb2.Operation]:
        # The return type is fine, but mypy isn't sophisticated enough to determine what's going on here.
        # In C++ this would require a dynamic_cast
        return self._DeleteSemanticGovernancePolicy(self._session, self._host, self._interceptor) # type: ignore

    @property
    def get_semantic_governance_policy(self) -> Callable[
            [semantic_governance_policy_service.GetSemanticGovernancePolicyRequest],
            semantic_governance_policy_service.SemanticGovernancePolicy]:
        # The return type is fine, but mypy isn't sophisticated enough to determine what's going on here.
        # In C++ this would require a dynamic_cast
        return self._GetSemanticGovernancePolicy(self._session, self._host, self._interceptor) # type: ignore

    @property
    def list_semantic_governance_policies(self) -> Callable[
            [semantic_governance_policy_service.ListSemanticGovernancePoliciesRequest],
            semantic_governance_policy_service.ListSemanticGovernancePoliciesResponse]:
        # The return type is fine, but mypy isn't sophisticated enough to determine what's going on here.
        # In C++ this would require a dynamic_cast
        return self._ListSemanticGovernancePolicies(self._session, self._host, self._interceptor) # type: ignore

    @property
    def update_semantic_governance_policy(self) -> Callable[
            [semantic_governance_policy_service.UpdateSemanticGovernancePolicyRequest],
            operations_pb2.Operation]:
        # The return type is fine, but mypy isn't sophisticated enough to determine what's going on here.
        # In C++ this would require a dynamic_cast
        return self._UpdateSemanticGovernancePolicy(self._session, self._host, self._interceptor) # type: ignore

    @property
    def kind(self) -> str:
        return "rest"

    def close(self):
        self._session.close()


__all__=(
    'SemanticGovernancePolicyServiceRestTransport',
)



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implements command to translate a SQL query."""

import functools
import json
import os
import typing

from apitools.base.py import encoding
from googlecloudsdk.api_lib.bq import util as api_util
from googlecloudsdk.api_lib.util import apis
from googlecloudsdk.api_lib.util import waiter
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.bq import command_utils
from googlecloudsdk.core import exceptions
from googlecloudsdk.core import log
from googlecloudsdk.core import properties
from googlecloudsdk.core import resources
from googlecloudsdk.core import yaml
from googlecloudsdk.core.console import console_io
from googlecloudsdk.core.util import files


@functools.cache
def _get_dialect_registry():
  """Gets dialect registry: input_dialects, output_dialects, dialect_pairs.

  The mapping is loaded from a JSON resource file.

  Returns:
    A dict with dialect registry.

  Raises:
    exceptions.Error: If the dialect registry fails to load.
  """
  try:
    file_path = os.path.join(os.path.dirname(__file__), 'dialect_registry.json')
    registry = json.loads(files.ReadFileContents(file_path))
    return registry
  except Exception as e:
    raise exceptions.Error(f'Failed to load dialect registry: {e}')


def _get_task_type(source_dialect: str, target_dialect: str) -> str:
  """Returns the migration task type based on the source dialect."""
  registry = _get_dialect_registry()
  source_dialect_map = {}
  for dialect in registry.get('input_dialects', []):
    legacy_name = dialect.get('legacy_batch_name')
    name = dialect.get('name')
    if legacy_name:
      if name:
        source_dialect_map[name.lower()] = legacy_name
      source_dialect_map[legacy_name.lower()] = legacy_name
    elif name:
      source_dialect_map[name.lower()] = name
  source_legacy_batch_name = source_dialect_map.get(
      source_dialect.lower(), source_dialect
  )
  target_dialect_map = {}
  for dialect in registry.get('output_dialects', []):
    legacy_name = dialect.get('legacy_batch_name')
    name = dialect.get('name')
    if legacy_name:
      if name:
        target_dialect_map[name.lower()] = legacy_name
      target_dialect_map[legacy_name.lower()] = legacy_name
    elif name:
      target_dialect_map[name.lower()] = name
  target_legacy_batch_name = target_dialect_map.get(
      target_dialect.lower(), target_dialect
  )
  task_type = (
      f'{source_legacy_batch_name}2{target_legacy_batch_name}_Translation'
  )
  valid_task_type = False
  for pair in registry['dialect_pairs']:
    if task_type in pair.get('legacy_batch_name', []):
      valid_task_type = True
      break
  if not valid_task_type:
    raise exceptions.Error(
        f'Translation from {source_dialect} to {target_dialect} is not'
        ' supported.'
    )
  return task_type


def _build_translation_details(
    messages,
    query,
    explanation_output_file=None,
    translation_config_files=None,
    metadata_gcs_uri=None,
    source_ddl_output_file=None,
):
  """Builds the translation details message for the migration task."""
  target_return_literals = ['sql/query.sql']
  target_types = ['sql']

  source_target_mapping = [
      messages.GoogleCloudBigqueryMigrationV2SourceTargetMapping(
          sourceSpec=messages.GoogleCloudBigqueryMigrationV2SourceSpec(
              literal=messages.GoogleCloudBigqueryMigrationV2Literal(
                  literalString=query, relativePath='query.sql'
              )
          )
      )
  ]

  if source_ddl_output_file:
    target_return_literals.append('source_sql_metadata_generation_suggestion/')
    if 'suggestion' not in target_types:
      target_types.append('suggestion')

  if explanation_output_file:
    target_return_literals.append(
        'translation_explanation_suggestion/query.sql'
    )
    if 'suggestion' not in target_types:
      target_types.append('suggestion')
    ai_config = (
        'translation_rules:\n  - suggestion_type: "TRANSLATION_EXPLANATION"'
    )
    source_target_mapping.append(
        messages.GoogleCloudBigqueryMigrationV2SourceTargetMapping(
            sourceSpec=messages.GoogleCloudBigqueryMigrationV2SourceSpec(
                literal=messages.GoogleCloudBigqueryMigrationV2Literal(
                    literalString=ai_config,
                    relativePath='gcloud_default_inline.ai_config.yaml',
                )
            )
        )
    )

  if translation_config_files:
    for config_file in translation_config_files:
      filename = os.path.basename(config_file)
      if not filename.endswith('.config.yaml'):
        raise exceptions.Error(
            f'Translation config file "{config_file}" must end with'
            ' ".config.yaml".'
        )
      content = files.ReadFileContents(config_file)
      source_target_mapping.append(
          messages.GoogleCloudBigqueryMigrationV2SourceTargetMapping(
              sourceSpec=messages.GoogleCloudBigqueryMigrationV2SourceSpec(
                  literal=messages.GoogleCloudBigqueryMigrationV2Literal(
                      literalString=content,
                      relativePath=filename,
                  )
              )
          )
      )

  if metadata_gcs_uri:
    source_target_mapping.append(
        messages.GoogleCloudBigqueryMigrationV2SourceTargetMapping(
            sourceSpec=messages.GoogleCloudBigqueryMigrationV2SourceSpec(
                gcsFilePath=metadata_gcs_uri
            )
        )
    )

  return messages.GoogleCloudBigqueryMigrationV2TranslationDetails(
      targetReturnLiterals=target_return_literals,
      targetTypes=target_types,
      sourceTargetMapping=source_target_mapping,
  )


def _build_migration_workflow(
    messages,
    query,
    source_dialect,
    target_dialect,
    explanation_output_file=None,
    translation_config_files=None,
    metadata_gcs_uri=None,
    source_ddl_output_file=None,
):
  """Builds the migration workflow message and returns it with the task type."""
  task_type = _get_task_type(source_dialect, target_dialect)
  translation_details = _build_translation_details(
      messages,
      query,
      explanation_output_file,
      translation_config_files,
      metadata_gcs_uri,
      source_ddl_output_file,
  )

  task = messages.GoogleCloudBigqueryMigrationV2MigrationTask(
      type=task_type, translationDetails=translation_details
  )

  workflow_tasks_value = messages.GoogleCloudBigqueryMigrationV2MigrationWorkflow.TasksValue(
      additionalProperties=[
          messages.GoogleCloudBigqueryMigrationV2MigrationWorkflow.TasksValue.AdditionalProperty(
              key='translation_task', value=task
          )
      ]
  )
  workflow = messages.GoogleCloudBigqueryMigrationV2MigrationWorkflow(
      tasks=workflow_tasks_value
  )
  return workflow, task_type


def _parse_task_result(wait_response):
  """Parses the workflow wait response into component task details."""
  translated_sql = None
  explanation = None
  generated_ddl = None
  result_task = None
  console_uri = ''
  report_log_messages = []

  if wait_response.tasks and wait_response.tasks.additionalProperties:
    translation_task_prop = next(
        (
            p
            for p in wait_response.tasks.additionalProperties
            if p.key == 'translation_task'
        ),
        None,
    )
    if translation_task_prop:
      result_task = translation_task_prop.value
    if (
        result_task
        and result_task.taskResult
        and result_task.taskResult.translationTaskResult
    ):
      literals = result_task.taskResult.translationTaskResult.translatedLiterals
      for literal in literals:
        if literal.relativePath == 'sql/query.sql':
          translated_sql = literal.literalString
        elif (
            literal.relativePath
            == 'translation_explanation_suggestion/query.sql'
        ):
          explanation = literal.literalString
        elif (
            literal.relativePath == 'source_sql_metadata_generation_suggestion/'
        ):
          generated_ddl = literal.literalString
      console_uri = result_task.taskResult.translationTaskResult.consoleUri
      report_log_messages = (
          result_task.taskResult.translationTaskResult.reportLogMessages
      )

  return (
      result_task,
      translated_sql,
      explanation,
      generated_ddl,
      console_uri,
      report_log_messages,
  )


def _format_log_result(
    wait_response, task_type, result_task, console_uri, report_log_messages
):
  """Formats the log result dictionary from the task components."""
  result = {
      'name': wait_response.name,
      'type': task_type,
      'state': wait_response.state.name if wait_response.state else '',
      'consoleUri': console_uri,
      'reportLogMessages': (
          [encoding.MessageToDict(msg) for msg in report_log_messages]
          if report_log_messages
          else []
      ),
  }

  if result_task:
    if result_task.processingError:
      result['processingError'] = encoding.MessageToDict(
          result_task.processingError
      )
    if result_task.resourceErrorDetails:
      result['resourceErrorDetails'] = [
          encoding.MessageToDict(msg)
          for msg in result_task.resourceErrorDetails
      ]

  return result


def _handle_task_failure(wait_response, result_task):
  """Checks the task state for failure and raises an exception if needed."""
  if result_task and result_task.state and result_task.state.name == 'FAILED':
    log.error('Workflow wait_response: %s', wait_response)
    if result_task.processingError:
      log.error('Task processing error: %s', result_task.processingError)
      error_details = []
      if (
          hasattr(result_task.processingError, 'message')
          and result_task.processingError.message
      ):
        error_details.append(result_task.processingError.message)
      if (
          hasattr(result_task.processingError, 'reason')
          and result_task.processingError.reason
      ):
        error_details.append(result_task.processingError.reason)
      error_str = ' '.join(error_details) if error_details else 'Unknown error'
      log.err.Print(f'Task processing error: {error_str}')
    if result_task.resourceErrorDetails:
      log.error('Task resource errors: %s', result_task.resourceErrorDetails)
    raise exceptions.Error('Failed to retrieve the translated SQL query.')


def _process_workflow_result(wait_response, task_type, args):
  """Processes the completed workflow and handles output/logging."""
  (
      result_task,
      translated_sql,
      explanation,
      generated_ddl,
      console_uri,
      report_log_messages,
  ) = _parse_task_result(wait_response)

  result = _format_log_result(
      wait_response, task_type, result_task, console_uri, report_log_messages
  )

  if args.translation_log_file:
    files.WriteFileContents(args.translation_log_file, yaml.dump(result))
    log.status.Print(
        f'Successfully saved translation logs to {args.translation_log_file}.'
    )

  _handle_task_failure(wait_response, result_task)

  if not translated_sql:
    raise exceptions.Error('Failed to retrieve the translated SQL query.')

  if args.output_file and args.output_file != '-':
    files.WriteFileContents(args.output_file, translated_sql)
    log.status.Print(
        f'Successfully translated and saved to {args.output_file}.'
    )
  else:
    log.out.Print('Translated query:')
    log.out.Print(translated_sql)

  if args.explanation_output_file and explanation:
    files.WriteFileContents(args.explanation_output_file, explanation)
    log.status.Print(
        'Successfully saved translation explanation to'
        f' {args.explanation_output_file}.'
    )

  if args.source_ddl_output_file and generated_ddl:
    files.WriteFileContents(args.source_ddl_output_file, generated_ddl)
    log.status.Print(
        f'Successfully saved source DDL to {args.source_ddl_output_file}.'
    )

  if args.translation_log_file:
    # Return None to avoid duplicating output in standard terminal output.
    return None

  log.out.Print('\nTranslation logs:')
  return result


@base.UniverseCompatible
@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class Translate(base.Command):
  """Translate a SQL query from a source dialect to BigQuery."""

  detailed_help = {
      'brief': 'Translate a SQL query from a source dialect to BigQuery.',
      'DESCRIPTION': """\
Translate a SQL query from a source dialect to BigQuery.

The command reads a SQL query from standard input (or an input file) and
prints the translated query to standard output (or an output file).
""",
      'EXAMPLES': """\
To translate a Snowflake query from stdin, run:

  $ echo 'SELECT * FROM test.my_table;' | {command} --source-dialect=SNOWFLAKE --location=us

To translate a Snowflake query from a file and save the output and logs to files, run:

  $ {command} \\
      --source-dialect=SNOWFLAKE \\
      --location=us \\
      --project=my-project \\
      --input-file=input.sql \\
      --output-file=output.sql \\
      --translation-log-file=translation_logs.yaml
""",
  }

  @staticmethod
  def Args(parser):
    parser.add_argument(
        '--source-dialect',
        help=(
            'Source dialect of the query. See supported dialects in'
            ' https://docs.cloud.google.com/bigquery/docs/batch-sql-translator#supported_sql_dialects'
        ),
        required=True,
    )
    parser.add_argument(
        '--target-dialect',
        help=(
            'Target dialect of the query. See supported dialects in'
            ' https://docs.cloud.google.com/bigquery/docs/batch-sql-translator#supported_sql_dialects'
        ),
        required=True,
    )
    parser.add_argument(
        '--location',
        help='Google Cloud Storage location to use for the translation.',
        required=True,
    )
    parser.add_argument(
        '--output-file',
        help=(
            'File to which to write the translated SQL. Writes to stdout if'
            " not provided or if set to ``-''."
        ),
        required=False,
    )
    parser.add_argument(
        '--input-file',
        help=(
            'File from which to read the SQL query to translate. Can be piped'
            ' through stdin as well.'
        ),
        required=False,
    )
    parser.add_argument(
        '--translation-log-file',
        help='File to which to write the translation logs.',
        required=False,
    )
    parser.add_argument(
        '--explanation-output-file',
        help='File to which to write the translation explanation.',
        required=False,
    )
    parser.add_argument(
        '--metadata-gcs-uri',
        help='Cloud Storage URI for the metadata zip file.',
        required=False,
    )
    parser.add_argument(
        '--translation-config-files',
        type=arg_parsers.ArgList(),
        metavar='CONFIG_FILE',
        help=(
            'A comma-separated list of paths to YAML configuration files. File'
            ' names must end with .config.yaml. See'
            ' https://docs.cloud.google.com/bigquery/docs/config-yaml-translation'
            ' for more details.'
        ),
        required=False,
    )
    parser.add_argument(
        '--source-ddl-output-file',
        help='File to which to write the generated source DDL.',
        required=False,
    )

  def Run(self, args):
    query = console_io.ReadFromFileOrStdin(args.input_file or '-', binary=False)

    if not query:
      raise exceptions.Error('No input query provided.')

    client = api_util.GetMigrationApiClient()
    client.additional_http_headers[api_util.TOOL_TAG_HEADER] = (
        api_util.GCLOUD_TOOL_TAG
    )
    messages = typing.cast(
        typing.Any, apis.GetMessagesModule('bigquerymigration', 'v2')
    )
    migration_service = client.projects_locations_workflows

    project = args.project or properties.VALUES.core.project.Get(required=True)
    location = args.location

    workflow, task_type = _build_migration_workflow(
        messages,
        query,
        args.source_dialect,
        args.target_dialect,
        explanation_output_file=args.explanation_output_file,
        translation_config_files=args.translation_config_files,
        metadata_gcs_uri=args.metadata_gcs_uri,
        source_ddl_output_file=args.source_ddl_output_file,
    )

    request_type = api_util.GetMigrationApiMessage(
        'BigquerymigrationProjectsLocationsWorkflowsCreateRequest'
    )
    request = request_type(
        parent=f'projects/{project}/locations/{location}',
        googleCloudBigqueryMigrationV2MigrationWorkflow=workflow,
    )

    response = migration_service.Create(request)

    workflow_ref = resources.REGISTRY.ParseRelativeName(
        response.name,
        collection='bigquerymigration.projects.locations.workflows',
    )
    poller = command_utils.BqMigrationWorkflowPoller(migration_service)

    wait_response = waiter.WaitFor(
        poller=poller,
        operation_ref=workflow_ref,
        message='Running translation workflow [{}]'.format(response.name),
    )

    return _process_workflow_result(wait_response, task_type, args)



# -*- coding: utf-8 -*- #
# Copyright 2016 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Flags and helpers for the firestore related commands."""


import argparse
import json
import string
import textwrap

from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import arg_parsers_usage_text
from googlecloudsdk.calliope import base as calliope_base

# Keys used in index field configuration.
FIELD_CONFIG_FIELD_PATH = 'field-path'
FIELD_CONFIG_ORDER = 'order'
FIELD_CONFIG_ARRAY_CONFIG = 'array-config'
FIELD_CONFIG_VECTOR_CONFIG = 'vector-config'
FIELD_CONFIG_SEARCH_CONFIG = 'search-config'
FIELD_CONFIG_DIMENSION = 'dimension'
FIELD_CONFIG_FLAT = 'flat'
FIELD_CONFIG_TEXT_SPEC = 'text-spec'
FIELD_CONFIG_INDEX_SPECS = 'index-specs'
FIELD_CONFIG_INDEX_TYPE = 'index-type'
FIELD_CONFIG_MATCH_TYPE = 'match-type'
FIELD_CONFIG_GEO_SPEC = 'geo-spec'
FIELD_CONFIG_GEO_JSON_INDEXING_DISABLED = 'geo-json-indexing-disabled'

# Keys used in search index options.
SEARCH_INDEX_OPTIONS_TEXT_LANGUAGE = 'text-language'
SEARCH_INDEX_OPTIONS_TEXT_LANGUAGE_OVERRIDE_FIELD_PATH = (
    'text-language-override-field-path'
)


class FlattenedAppendAction(argparse.Action):
  """Custom argparse action that flattens parsed lists on append.

  This avoids producing nested list-of-lists (e.g., [[obj1], [obj2]]) when
  ArgObject.repeated=True is used in conjunction with parser action='append'.
  """

  def __call__(self, parser, namespace, values, option_string=None):
    items = getattr(namespace, self.dest, None) or []
    if isinstance(values, list):
      items.extend(values)
    else:
      items.append(values)
    setattr(namespace, self.dest, items)


def AddCollectionGroupIdsFlag(parser):
  """Adds flag for collection group ids to the given parser.

  Args:
    parser: The argparse parser.
  """
  parser.add_argument(
      '--collection-ids',
      metavar='COLLECTION_GROUP_IDS',
      type=arg_parsers.ArgList(),
      help=textwrap.dedent("""\
          List specifying which collection groups will be included in the operation.
          When omitted, all collection groups are included.

          For example, to operate on only the `customers` and `orders`
          collections groups:

            $ {command} --collection-ids='customers','orders'
          """),
  )


def AddDatabaseIdFlag(parser, required=False, hidden=False):
  """Adds flag for database id to the given parser.

  Args:
    parser: The argparse parser.
    required: Whether the flag must be set for running the command, a bool.
    hidden: Whether the flag is hidden, a bool.
  """
  if not required:
    helper_text = textwrap.dedent("""\
        The database to operate on. The default value is `(default)`.

        For example, to operate on database `foo`:

          $ {command} --database='foo'
        """)
  else:
    helper_text = textwrap.dedent("""\
        The database to operate on.

        For example, to operate on database `foo`:

          $ {command} --database='foo'
        """)
  parser.add_argument(
      '--database',
      metavar='DATABASE',
      type=str,
      default='(default)' if not required else None,
      required=required,
      hidden=hidden,
      help=helper_text,
  )


def AddNamespaceIdsFlag(parser):
  """Adds flag for namespace ids to the given parser."""
  parser.add_argument(
      '--namespace-ids',
      metavar='NAMESPACE_IDS',
      type=arg_parsers.ArgList(),
      help=textwrap.dedent("""\
          List specifying which namespaces will be included in the operation.
          When omitted, all namespaces are included.

          This is only supported for Datastore Mode databases.

          For example, to operate on only the `customers` and `orders` namespaces:

            $ {command} --namespaces-ids='customers','orders'
          """),
  )


def AddSnapshotTimeFlag(parser):
  """Adds flag for snapshot time to the given parser.

  Args:
    parser: The argparse parser.
  """
  parser.add_argument(
      '--snapshot-time',
      metavar='SNAPSHOT_TIME',
      type=str,
      default=None,
      required=False,
      help=textwrap.dedent("""\
          The version of the database to export.

          The timestamp must be in the past, rounded to the minute and not older
          than `earliestVersionTime`. If specified, then the exported documents will
          represent a consistent view of the database at the provided time.
          Otherwise, there are no guarantees about the consistency of the exported
          documents.

          For example, to operate on snapshot time `2023-05-26T10:20:00.00Z`:

            $ {command} --snapshot-time='2023-05-26T10:20:00.00Z'
          """),
  )


def AddLocationFlag(
    parser, required=False, hidden=False, suggestion_aliases=None
):
  """Adds flag for location to the given parser.

  Args:
    parser: The argparse parser.
    required: Whether the flag must be set for running the command, a bool.
    hidden: Whether the flag is hidden in document. a bool.
    suggestion_aliases: A list of flag name aliases. A list of string.
  """
  parser.add_argument(
      '--location',
      metavar='LOCATION',
      required=required,
      hidden=hidden,
      type=str,
      suggestion_aliases=suggestion_aliases,
      help=textwrap.dedent("""\
          The location to operate on. Available locations are listed at
          https://cloud.google.com/firestore/docs/locations.

          For example, to operate on location `us-east1`:

            $ {command} --location='us-east1'
          """),
  )


def AddBackupFlag(parser):
  """Adds flag for backup to the given parser.

  Args:
    parser: The argparse parser.
  """
  parser.add_argument(
      '--backup',
      metavar='BACKUP',
      required=True,
      type=str,
      help=textwrap.dedent("""\
          The backup to operate on.

          For example, to operate on backup `cf9f748a-7980-4703-b1a1-d1ffff591db0`:

            $ {command} --backup='cf9f748a-7980-4703-b1a1-d1ffff591db0'
          """),
  )


def AddBackupScheduleFlag(parser):
  """Adds flag for backup schedule id to the given parser.

  Args:
    parser: The argparse parser.
  """
  parser.add_argument(
      '--backup-schedule',
      metavar='BACKUP_SCHEDULE',
      required=True,
      type=str,
      help=textwrap.dedent("""\
          The backup schedule to operate on.

          For example, to operate on backup schedule `091a49a0-223f-4c98-8c69-a284abbdb26b`:

            $ {command} --backup-schedule='091a49a0-223f-4c98-8c69-a284abbdb26b'
          """),
  )


def AddRetentionFlag(parser, required=False):
  """Adds flag for retention to the given parser.

  Args:
    parser: The argparse parser.
    required: Whether the flag must be set for running the command, a bool.
  """
  parser.add_argument(
      '--retention',
      metavar='RETENTION',
      required=required,
      type=arg_parsers.Duration(),
      help=textwrap.dedent("""\
          The rention of the backup. At what relative time in the future,
          compared to the creation time of the backup should the backup be
          deleted, i.e. keep backups for 7 days.

          For example, to set retention as 7 days.

          $ {command} --retention=7d
          """),
  )


def AddRecurrenceFlag(parser):
  """Adds flag for recurrence to the given parser.

  Args:
    parser: The argparse parser.
  """
  group = parser.add_group(
      help=textwrap.dedent("""\
          Recurrence settings of a backup schedule.
          """),
      required=True,
  )
  help_text = textwrap.dedent("""\
      The recurrence settings of a backup schedule.

      Currently only daily and weekly backup schedules are supported.

      When a weekly backup schedule is created, day-of-week is needed.

      For example, to create a weekly backup schedule which creates backups on
      Monday.

        $ {command} --recurrence=weekly --day-of-week=MON
  """)
  group.add_argument('--recurrence', type=str, help=help_text, required=True)

  help_text = textwrap.dedent("""\
     The day of week (UTC time zone) of when backups are created.

      The available values are: `MON`, `TUE`, `WED`, `THU`, `FRI`, `SAT`,`SUN`.
      Values are case insensitive.

      This is required when creating a weekly backup schedule.
  """)
  group.add_argument(
      '--day-of-week',
      choices=arg_parsers.DayOfWeek.DAYS,
      type=arg_parsers.DayOfWeek.Parse,
      help=help_text,
      required=False,
  )


def AddEncryptionConfigGroup(parser, source_type):
  """Adds flags for the database's encryption configuration to the given parser.

  Args:
    parser: The argparse parser.
    source_type: "backup" if a restore; "database" if a clone
  """
  encryption_config = parser.add_argument_group(
      required=False,
      help=textwrap.dedent(string.Template("""\
            The encryption configuration of the new database being created from the $source_type.
            If not specified, the same encryption settings as the $source_type will be used.

            To create a CMEK-enabled database:

              $$ {command} --encryption-type=customer-managed-encryption --kms-key-name=projects/PROJECT_ID/locations/LOCATION_ID/keyRings/KEY_RING_ID/cryptoKeys/CRYPTO_KEY_ID

            To create a Google-default-encrypted database:

              $$ {command} --encryption-type=google-default-encryption

            To create a database using the same encryption settings as the $source_type:

              $$ {command} --encryption-type=use-source-encryption
            """).substitute(source_type=source_type)),
  )
  encryption_config.add_argument(
      '--encryption-type',
      metavar='ENCRYPTION_TYPE',
      type=str,
      required=True,
      choices=[
          'use-source-encryption',
          'customer-managed-encryption',
          'google-default-encryption',
      ],
      help=textwrap.dedent("""\
          The encryption type of the destination database.
          """),
  )
  AddKmsKeyNameFlag(
      encryption_config,
      'This flag must only be specified when encryption-type is'
      ' `customer-managed-encryption`.',
  )


def AddConcurrencyModeFlag(parser):
  """Adds flag for concurrency mode to the given parser.

  Args:
    parser: The argparse parser.
  """
  parser.add_argument(
      '--concurrency-mode',
      metavar='CONCURRENCY_MODE',
      type=str,
      choices=['optimistic', 'pessimistic'],
      help=textwrap.dedent("""\
          The concurrency control mode to use for this database.

          When not specified, Firestore will pick a default concurrency mode
          based on the database edition.
          """),
  )


def AddKmsKeyNameFlag(parser, additional_help_text=None):
  """Adds flag for KMS Key Name to the given parser.

  Args:
    parser: The argparse parser.
    additional_help_text: Additional help text to be added to the flag.
  """

  help_text = textwrap.dedent("""\
      The resource ID of a Cloud KMS key. If set, the database created will be a Customer-Managed Encryption Key (CMEK) database encrypted with this key.
      This feature is allowlist only in initial launch.

      Only a key in the same location as this database is allowed to be used for encryption.
      For Firestore's nam5 multi-region, this corresponds to Cloud KMS location us.
      For Firestore's eur3 multi-region, this corresponds to Cloud KMS location europe.
      See https://cloud.google.com/kms/docs/locations.

      This value should be the KMS key resource ID in the format of `projects/{project_id}/locations/{kms_location}/keyRings/{key_ring}/cryptoKeys/{crypto_key}`.
      How to retrieve this resource ID is listed at https://cloud.google.com/kms/docs/getting-resource-ids#getting_the_id_for_a_key_and_version.
    """)
  if additional_help_text:
    help_text = help_text + '\n\n' + additional_help_text

  parser.add_argument(
      '--kms-key-name',
      metavar='KMS_KEY_NAME',
      type=str,
      required=False,
      default=None,
      help=help_text,
  )


def AddDestinationDatabase(parser, action_name, source_type):
  parser.add_argument(
      '--destination-database',
      metavar='DESTINATION_DATABASE',
      type=str,
      required=True,
      help=textwrap.dedent(f"""\
          Destination database to {action_name} to. Destination database will be created in the same location as the source {source_type}.

          This value should be 4-63 characters. Valid characters are /[a-z][0-9]-/
          with first character a letter and the last a letter or a number. Must
          not be UUID-like /[0-9a-f]{8}(-[0-9a-f]{4}){3}-[0-9a-f]{12}/.

          Using "(default)" database ID is also allowed.

          For example, to {action_name} to database `testdb`:

          $ {{command}} --destination-database=testdb
          """),
  )


def AddTags(parser, resource_type):
  """Adds the --tags flag to the given parser.

  Args:
    parser: The parser to add the flag to.
    resource_type: The resource type to use in the help text (e.g. 'database').
  """
  parser.add_argument(
      '--tags',
      metavar='KEY=VALUE',
      type=arg_parsers.ArgDict(),
      default=None,
      help=textwrap.dedent(f"""\
          Tags to attach to the destination {resource_type}. Example: --tags=key1=value1,key2=value2

          For example, to attach tags to a {resource_type}:

          $ --tags=key1=value1,key2=value2
          """),
  )


def AddUserCredsIdArg(parser):
  """Adds positional arg for user creds id to the given parser.

  Args:
    parser: The argparse parser.
  """
  parser.add_argument(
      'user_creds',
      metavar='USER_CREDS',
      type=str,
      help="""
      The user creds to operate on.

      For example, to operate on user creds `creds-name-1`:

        $ {command} creds-name-1
      """,
  )


def _ChoiceType(choices):
  """Adapts calliope_base.ChoiceArgument behavior for use with nested arg_parse.ArgObject.

  Args:
    choices: [str], A list of valid flag values.

  Returns:
    A callable that sanitizes and validates a user provided flag value against
    those choices. This callable matches ChoiceArgument behavior.
  """
  # Validate the choice list with calliope_base.
  _ = calliope_base.ChoiceArgument('unused', choices=choices)

  def Parse(val):
    val_sanitized = calliope_base.SanitizeChoices([val])[0]
    if val_sanitized not in choices:
      raise arg_parsers.ArgumentTypeError(
          f'Invalid choice: {val}. Valid choices are: [{", ".join(choices)}].'
      )
    return val_sanitized

  return Parse


_FIELD_PATH_SPEC = arg_parsers.ArgObject(
    value_type=str,
    help_text='specifies the field path (e.g. address.city).',
)

_ARRAY_CONFIG_SPEC = arg_parsers.ArgObject(
    value_type=_ChoiceType(['contains']),
    help_text=(
        'Specifies the configuration for an array field. The only '
        "valid option is 'contains'. Exactly one of 'order', "
        "'array-config', or 'vector-config' must be specified."
    ),
)

_ORDER_SPEC = arg_parsers.ArgObject(
    value_type=_ChoiceType(['ascending', 'descending', 'order-unspecified']),
    help_text=(
        "Specifies the order. Valid options are 'ascending', "
        "'descending'. Exactly one of 'order', 'array-config', or "
        "'vector-config' must be specified."
    ),
)

_VECTOR_CONFIG_SPEC = arg_parsers.ArgObject(
    help_text=(
        'Specifies the configuration for a vector field. Exactly one of '
        "'order', 'array-config', or 'vector-config' must be specified."
    ),
    spec={
        FIELD_CONFIG_DIMENSION: arg_parsers.ArgObject(
            value_type=int,
        ),
        FIELD_CONFIG_FLAT: arg_parsers.ArgObject(
            spec={},
        ),
    },
    required_keys=(FIELD_CONFIG_DIMENSION,),
)

_SEARCH_CONFIG_SPEC = arg_parsers.ArgObject(
    help_text=textwrap.dedent("""\
        Specifies the configuration for a search field. An index definition
        must contain either only 'search-config' fields or only non
        'search-config' fields.

        The following shorthand aliases are supported instead of a full 'search-config':
          * `TEXT_TOKENIZED_MATCH_GLOBALLY`: Tokenized text search with global matching.
          * `GEO_POINT`: Geo search.

        Examples:

        With alias:

            --field-config=field-path=title,search-config=TEXT_TOKENIZED_MATCH_GLOBALLY

        Text search:

            --field-config=field-path=title,search-config='{"text-spec": {"index-specs": [{"index-type": "tokenized", "match-type": "match-globally"}]}}'

        Geo search:

            --field-config=field-path=location,search-config='{"geo-spec": {"geo-json-indexing-disabled": true}}'

        With file:

            --field-config=field-path=text,search-config='/path/to/configs/search-config.json'

        For complex configurations, it is recommended to use a file.
        """),
    spec={
        FIELD_CONFIG_TEXT_SPEC: arg_parsers.ArgObject(
            help_text=(
                'Optional. The specification for building a text search '
                'index for a field.'
            ),
            spec={
                FIELD_CONFIG_INDEX_SPECS: arg_parsers.ArgObject(
                    help_text=(
                        'Optional. Array of specifications for how the '
                        'field should be indexed.'
                    ),
                    repeated=True,
                    spec={
                        FIELD_CONFIG_INDEX_TYPE: arg_parsers.ArgObject(
                            value_type=_ChoiceType(['tokenized']),
                            help_text=textwrap.dedent("""\
                                Required. How to index the text field value. Valid options are:
                                  * `tokenized`: Field values are tokenized.
                                """),
                        ),
                        FIELD_CONFIG_MATCH_TYPE: arg_parsers.ArgObject(
                            value_type=_ChoiceType(['match-globally']),
                            help_text=textwrap.dedent("""\
                                Required. How to match the text field value. Valid options are:
                                  * `match-globally`: Match on any indexed field.
                                """),
                        ),
                    },
                ),
            },
        ),
        FIELD_CONFIG_GEO_SPEC: arg_parsers.ArgObject(
            help_text=(
                'Optional. The specification for building a geo search '
                'index for a field.'
            ),
            spec={
                FIELD_CONFIG_GEO_JSON_INDEXING_DISABLED: arg_parsers.ArgObject(
                    value_type=bool,
                    help_text=(
                        'Optional. Disables geoJSON indexing for the '
                        'field. By default, geoJSON points are indexed.'
                    ),
                ),
            },
        ),
    },
)


_SEARCH_CONFIG_ALIASES = {
    'TEXT_TOKENIZED_MATCH_GLOBALLY': json.dumps({
        'text-spec': {
            'index-specs': [{
                'index-type': 'tokenized',
                'match-type': 'match-globally',
            }]
        }
    }),
    'GEO_POINT': json.dumps({'geo-spec': {}}),
}


# Inheriting from DefaultArgTypeWrapper allows us to support aliases without
# sacrificing the generated help text.
class _SearchConfigTypeWrapper(arg_parsers_usage_text.DefaultArgTypeWrapper):
  """Parses search config, expanding aliases if present."""

  def __init__(self):
    # Pass the search config spec to the parent class so it gets saved to
    # self.arg_type. This allows Calliope to extract the usage help text.
    super().__init__(_SEARCH_CONFIG_SPEC)

  def __call__(self, val):
    if isinstance(val, str):
      val_upper = val.upper()
      if val_upper in _SEARCH_CONFIG_ALIASES:
        val = _SEARCH_CONFIG_ALIASES[val_upper]

    return self.arg_type(val)


def AddFieldConfigFlag(parser, is_search_released):
  """Adds the repeated --field-config flag to the given parser.

  Args:
    parser: The argparse parser.
    is_search_released: Whether search is released in gcloud.
  """
  field_config_spec = {
      FIELD_CONFIG_FIELD_PATH: _FIELD_PATH_SPEC,
      FIELD_CONFIG_ARRAY_CONFIG: _ARRAY_CONFIG_SPEC,
      FIELD_CONFIG_ORDER: _ORDER_SPEC,
      FIELD_CONFIG_VECTOR_CONFIG: _VECTOR_CONFIG_SPEC,
  } | (
      {FIELD_CONFIG_SEARCH_CONFIG: _SearchConfigTypeWrapper()}
      if is_search_released
      else {}
  )

  help_text = 'Configuration for an index field.'

  parser.add_argument(
      '--field-config',
      type=arg_parsers.ArgObject(
          help_text=help_text,
          spec=field_config_spec,
          required_keys=(FIELD_CONFIG_FIELD_PATH,),
          enable_shorthand=True,
          disable_key_description=False,
          repeated=True,
      ),
      required=True,
      action=FlattenedAppendAction,
      help=help_text,
  )


def AddQueryScopeFlag(parser):
  """Adds the --query-scope flag to the given parser.

  Args:
    parser: The argparse parser.
  """
  calliope_base.ChoiceArgument(
      '--query-scope',
      choices=['collection', 'collection-group', 'collection-recursive'],
      default='collection',
      help_str='Query scope the index applies to.',
  ).AddToParser(parser)


def AddApiScopeFlag(parser):
  """Adds the --api-scope flag to the given parser.

  Args:
    parser: The argparse parser.
  """
  calliope_base.ChoiceArgument(
      '--api-scope',
      choices=['any-api', 'datastore-mode-api', 'mongodb-compatible-api'],
      default='any-api',
      help_str='Api scope the index applies to.',
  ).AddToParser(parser)


def AddDensityFlag(parser):
  """Adds the --density flag to the given parser.

  Args:
    parser: The argparse parser.
  """
  calliope_base.ChoiceArgument(
      '--density',
      choices=['dense', 'density-unspecified', 'sparse-all', 'sparse-any'],
      default=None,
      help_str='Density of the index.',
  ).AddToParser(parser)


def AddMultikeyFlag(parser):
  """Adds the --multikey flag to the given parser.

  Args:
    parser: The argparse parser.
  """
  parser.add_argument(
      '--multikey',
      action='store_true',
      help=textwrap.dedent("""\
          Optional. Whether the index is multikey. By default, the index
          is not multikey. For non-multikey indexes, none of the paths in the
          index definition reach or traverse an array, except via an explicit
          array index. For multikey indexes, at most one of the paths in the index
          definition reach or traverse an array, except via an explicit array
          index. Violations will result in errors. Note this field only applies to
          index with 'mongodb-compatible-api' ApiScope.
      """),
  )


def AddUniqueFlag(parser):
  """Adds the --unique flag to the given parser.

  Args:
    parser: The argparse parser.
  """
  parser.add_argument(
      '--unique',
      action='store_true',
      help=textwrap.dedent("""\
          Optional. Whether it is an unique index. Unique index ensures all values for
          the indexed field(s) are unique across documents.
      """),
  )


_SEARCH_INDEX_OPTIONS_SPEC = arg_parsers.ArgObject(
    help_text='Optional. Configuration options for search indexes.',
    spec={
        SEARCH_INDEX_OPTIONS_TEXT_LANGUAGE: arg_parsers.ArgObject(
            value_type=str,
            help_text=(
                'Optional. The language to use for text search '
                'indexes. Used as the default language if not '
                'overridden at the document level by specifying the '
                "'text-language-override-field-path'. The language is "
                'specified as a BCP 47 language code. For indexes '
                "with 'mongodb-compatible-api' ApiScope: If "
                'unspecified, the default language is English. For '
                "indexes with 'any-api' ApiScope: If unspecified, "
                'the default behavior is autodetect.'
            ),
        ),
        SEARCH_INDEX_OPTIONS_TEXT_LANGUAGE_OVERRIDE_FIELD_PATH: arg_parsers.ArgObject(
            value_type=str,
            help_text=(
                'Optional. The field in the document that specifies'
                ' which language to use for that specific document. If'
                ' unspecified, the language is taken from the'
                " 'language' document field if it exists or from"
                " 'text-language' if it does not."
            ),
        ),
    },
    disable_key_description=False,
)


def AddSearchIndexOptionsFlag(parser, is_search_released):
  """Adds the --search-index-options flag to the given parser.

  Args:
    parser: The argparse parser.
    is_search_released: Whether search is released in gcloud.
  """
  if is_search_released:
    search_index_help = 'Optional. Configuration options for search indexes.'
    parser.add_argument(
        '--search-index-options',
        type=_SEARCH_INDEX_OPTIONS_SPEC,
        required=False,
        help=search_index_help,
    )



# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for listing available worker pool revisions."""

from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.run import commands
from googlecloudsdk.command_lib.run import connection_context
from googlecloudsdk.command_lib.run import flags
from googlecloudsdk.command_lib.run import platforms
from googlecloudsdk.command_lib.run import pretty_print
from googlecloudsdk.command_lib.run import resource_args
from googlecloudsdk.command_lib.run import serverless_operations
from googlecloudsdk.command_lib.util.concepts import concept_parsers
from googlecloudsdk.command_lib.util.concepts import presentation_specs
from googlecloudsdk.core import log


@base.UniverseCompatible
@base.ReleaseTracks(base.ReleaseTrack.GA)
class List(commands.List):
  """List available worker pool revisions."""

  detailed_help = {
      'DESCRIPTION': (
          """\
          {description}
          """
      ),
      'EXAMPLES': (
          """\
          To list all revisions in a worker pool `foo`:

              $ {command} --worker-pool=foo
          """
      ),
  }

  @classmethod
  def CommonArgs(cls, parser):
    worker_pool_presentation = presentation_specs.ResourcePresentationSpec(
        '--namespace',
        resource_args.GetNamespaceResourceSpec(),
        'Namespace to list revisions in.',
        required=True,
        prefixes=False,
        hidden=True,
    )
    concept_parsers.ConceptParser([worker_pool_presentation]).AddToParser(
        parser
    )

    flags.AddWorkerPoolFlag(parser)
    flags.AddRegionArg(parser)

    parser.display_info.AddFormat(
        'table('
        '{ready_column},'
        'name:label=REVISION,'
        'active.yesno(yes="yes", no=""),'
        'worker_pool_name:label=WORKER_POOL:sort=1,'
        'creation_timestamp.date("%Y-%m-%d %H:%M:%S %Z"):'
        'label=DEPLOYED:sort=2:reverse,'
        'author:label="DEPLOYED BY"):({alias})'.format(
            ready_column=pretty_print.READY_COLUMN,
            alias=commands.SATISFIES_PZS_ALIAS,
        )
    )

  @classmethod
  def Args(cls, parser):
    cls.CommonArgs(parser)

  def _FilterServiceRevisions(self, revisions):
    """Filters out revisions that are service revisions.

    Per discussion with jmahood@, we want to make sure that all resources are
    self-contained, so none of the describe/list commands should mix the
    resource type.

    Args:
      revisions: List of revisions to filter.

    Returns:
      List of revisions that are worker pool revisions.
    """
    return list(filter(lambda rev: rev.worker_pool_name is not None, revisions))

  def Run(self, args):
    """List available revisions."""
    label_selector = None
    worker_pool_name = args.worker_pool
    conn_context = connection_context.GetConnectionContext(
        args, flags.Product.RUN, self.ReleaseTrack()
    )
    namespace_ref = args.CONCEPTS.namespace.Parse()
    with serverless_operations.Connect(conn_context) as client:
      self.SetCompleteApiEndpoint(conn_context.endpoint)
      if platforms.GetPlatform() != platforms.PLATFORM_MANAGED:
        location_msg = ' in [{}]'.format(conn_context.cluster_location)
        log.status.Print(
            'For cluster [{cluster}]{zone}:'.format(
                cluster=conn_context.cluster_name,
                zone=location_msg if conn_context.cluster_location else '',
            )
        )
      if worker_pool_name is not None:
        label_selector = 'run.googleapis.com/workerPool = {}'.format(
            worker_pool_name
        )
      for rev in self._FilterServiceRevisions(
          client.ListRevisions(
              namespace_ref, label_selector, args.limit, args.page_size
          )
      ):
        yield rev


@base.ReleaseTracks(base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA)
@base.RegionalEndpointsSupported
class BetaList(List):
  """List available worker pool revisions."""
